import sys
import serial
import time
import RPi.GPIO as GPIO
from time import sleep
from log import Logging
from com import Arduino


class Temp:
    def __init__(self, kp, ki, kd, set_point, simple_time=1, windup_limit=20):
        self.kp = kp
        self.ki = ki
        self.kd = kd
        self.set_point = set_point
        self.simple_time = simple_time
        self.windup_limit = windup_limit

        self.previous_error = 0
        self.integral = 0
        self.last_time = time.time()
        # for grinder

        time.sleep(1)


    def compute(self, feedback):

        current_time = time.time()
        elapsed_time = current_time - self.last_time

        if elapsed_time < self.simple_time:
           return 0

        error = self.set_point - feedback

        P = self.kp * error

        self.integral += error * elapsed_time
        if self.integral > self.windup_limit:
            self.integral = self.windup_limit
        elif self.integral < -self.windup_limit:
            self.integral = -self.windup_limit
        I = self.ki * self.integral

        derivative = (error - self.previous_error) / elapsed_time
        D = self.kd * derivative

        self.previous_error = error
        self.last_time = current_time

        return P + I + D

# the class is for devices
class Machine:

    def __init__(self, ID, price: int, coffee_powder: float, coffee: float,
                 sugar: float, hotMilk: float, frothMilk: float, chocolate_powder: float, cocoa_powder: float,
                 out_pin_info: str, per_sec: str):

        # initialize input variables
        self.ID = ID
        self.powder = 0.0     # for grinder
        self.price = price
        self.coffee_powder = coffee_powder
        self.coffee = coffee
        self.sugar = sugar
        self.hotMilk = hotMilk
        self.frothMilk = frothMilk
        self.chocolate_powder = chocolate_powder
        self.cocoa_powder = cocoa_powder

        # init static value
        self.modCoffee = None
        self.modCoffeePowder = None
        self.none = None
        self.per_machine = True

        self.error = False

        # for file.txt path like string
        self.out_pin_info = out_pin_info
        self.per_sec = per_sec


        # create dect for pins
        self.oPin = {}
        self.pSec = {}

        # split and append to dect
        with open(self.out_pin_info, 'r', encoding='utf8') as file:
            for line in file:
                key, val = line.strip().split(':')
                val = int(val)
                self.oPin[key] = val
        with open(self.per_sec, 'r', encoding='utf8') as file:
            for line in file:
                key, val = line.strip().split(':')
                val = float(val)
                self.pSec[key] = val
        self.pin_heater = self.oPin['pinHeater']
        self.pin_pump = self.oPin['pinPump']

        #  initialize pins in #GPIO

        GPIO.setmode(GPIO.BCM)
        GPIO.setwarnings(False)
        for val in self.oPin.values() :
            GPIO.setup(val, GPIO.OUT)
            GPIO.output(val, True)
            


        self.log_machine = Logging("machine")


        # arduino com connection
        self.arduino = Arduino()

        self.log_machine.get_logger(self.ID, "machine initialized", 1)
    # def for off all outputs
    def off(self):
        for i in self.oPin.values():
            GPIO.output(i, True)

    # def for check devices are in the right place
    def check_sensor(self, sen:str):
        if self.per_machine :
            try :
                global dat
                data = self.arduino.main(sen)
                list = ["2", "3", "4"]
                if sen in list and data is not None :
                    try:
                        dat = int(data)
                        self.none = 0
                    except Exception:
                        dat = 1
                        self.none += 1
                if self.none >= 5:
                    raise Exception
            except Exception as e:
                self.off()
                dat = None
                self.error = True
                self.per_machine = False
                self.log_machine.get_logger(self.ID, f"valve failed : sensor response is None", 3)

        return dat

    def get_temp(self):
        if self.per_machine :
            try :
                sen = "1"
                data = self.arduino.main(sen)
                if data is not None :
                    try :
                        dat = data.split()
                        dat1 = float(dat[0])
                        dat2 = float(dat[1])
                        self.none = 0
                    except Exception :
                        dat1 = 250
                        dat2 = 250
                        self.none += 1
                else :
                    dat1 = 200
                    dat2 = 200
                    self.none += 1
                if  self.none >= 5:
                    raise Exception
                return dat1, dat2
            except Exception as e :
                dat1 = None
                dat2 = None
                self.off()
                self.error = True
                self.per_machine = False
                self.log_machine.get_logger(self.ID, f"get temp failed : sensor response is None", 3)
                return dat1, dat2
    # for valve device
    def valve(self, state: int):
        if self.per_machine :
            try :
                # can be a number between 0 and 7
                pin = self.oPin['pinValve']
                self.arduino = Arduino()
                list = [0, 1, 3, 5, 7]
                if state in list:
                    GPIO.output(pin, False)
                    sleep(3)
                    t = 0
                    mic = 1
                    while True :
                        mic = self.check_sensor("2")
                        if mic == 0:
                            t += 1
                        if mic == 1:
                            t = 0
                        if t == 16:
                            break
                        if mic is None :
                            raise Exception


                    GPIO.output(pin, True)
                    sleep(0.1)
                    t = 0
                    mic = 1
                    GPIO.output(pin, False)
                    for i in range(state):
                        while True :
                            mic = self.check_sensor("2")
                            if mic == 0:
                                t += 1
                            if mic == 1:
                                t = 0
                            if t == 9:
                                t = 0
                                break
                            if mic is None:
                                raise Exception
                    GPIO.output(pin, True)

                list2 = [10, 20, 30, 40, 50]
                if state in list2 :
                    for i in range(int(state/10)):
                        t = 0
                        mic = 1
                        GPIO.output(pin, False)
                        while True :
                            mic = self.check_sensor("2")
                            if mic == 0:
                                t += 1
                            if mic == 1:
                                t = 0
                            if t == 9:
                                t = 0
                                break
                            if mic is None:
                                raise Exception
                        GPIO.output(pin, True)
                sleep(0.5)

                self.log_machine.get_logger(self.ID, "Valve done", 1)
            except Exception as e :
                self.off()
                self.error = True
                self.per_machine = False
                self.log_machine.get_logger(self.ID, f"valve failed : {e}", 3)
    # for driver device
    def derive(self, state: int):
        if self.per_machine :
            try :
                # add number of pins
                pinW = self.oPin['pinDriverWork']
                pinH = self.oPin['pinDriverHome']
                self.arduino = Arduino()
                # check position
                mich = 1
                micw = 1
                t= 0
                if state == 0:
                    GPIO.output(pinH, False)
                    mich = 1
                    mich = self.check_sensor("4")
                    sleep(0.5)
                    while True :
                        mich = self.check_sensor("4")     # "4" for home
                        t += 0.01
                        if mich == 0:
                            GPIO.output(pinH, True)
                            break
                        if t >= 6 :
                            break
                        if mich is None:
                            raise Exception
                if state == 1:
                    GPIO.output(pinW, False)
                    sleep(0.5)
                    while True:
                        micw = self.check_sensor("3") # "3" for work
                        t += 0.01
                        if micw == 0:
                            GPIO.output(pinW, True)
                            break
                        if t >= 6 :
                            break
                        if mich is None:
                            raise Exception
                GPIO.output(pinW, True)
                GPIO.output(pinH, True)
                sleep(0.5)
                self.log_machine.get_logger(self.ID,"Deriver done", 1)
            except Exception as e :
                self.off()
                self.error = True
                self.per_machine = False
                self.log_machine.get_logger(self.ID, f"valve failed : {e}", 3)

    # heating water
    def hot_water(self, temp_water: int, temp_body: int, ml: float):
        global T_water
        global T_body

        if self.per_machine :
            try :
                if ml > 0 :
                    self.per_sec = self.pSec['pump']
                    # set temperature given by class Temp
                    T = Temp(kp=1, ki=0, kd=0.1, set_point=temp_body +20, simple_time=0.31)


                    # set some statics variables
                    t = 0
                    sleep(2)
                    time = ml * self.per_sec
                    t_water = 0

                    # while loop for checking evry 0.3 seconds
                    # second loop extraction coffee
                    while True:
                        # receive water temperature from arduino
                        sleep(0.3)

                        T_body, T_water = self.get_temp()
                        if T_body or T_water is None :
                            raise Exception

                        current_temp = T_water
                        # put the body temperature in Temp class to compute PID control result
                        control_output = T.compute(current_temp)
                        # check to turn on  or turned off heater

                        if control_output > 0:
                            GPIO.output(self.pin_heater, False)
                        else:
                            GPIO.output(self.pin_heater, True)

                        t += 0.31

                        GPIO.output(self.pin_pump, True)

                        if T_water >= temp_water or T_body >= temp_body :
                            GPIO.output(self.pin_pump, False)
                            sleep(time)
                            GPIO.output(self.pin_pump, True)
                            sleep(0.5)
                            t_water += time

                        if t_water >= time :
                            break



                    # shut down pump and heater
                    GPIO.output(self.pin_pump, True)
                    GPIO.output(self.pin_heater, True)
                self.log_machine.get_logger(self.ID, f" Hot water Done!! in water temp {T_water}", 1)
            except Exception as e :
                self.off()
                self.error = True
                self.per_machine = False
                self.log_machine.get_logger(self.ID, f"Hot water failed : {e}", 3)


    # hot milk
    def hot_milk(self, temp: int, ml: float):
        global T_water
        global T_body
        global per
        per = True
        if self.per_machine :
            try :
                if ml > 0 :
                    self.per_sec = self.pSec['pump']

                    # set temperature given by class Temp
                    T = Temp(kp=1, ki=0.1, kd=0.8, set_point=temp+20, simple_time=0.31)

                    # set some statics variables
                    t = 0
                    sleep(2)
                    time = ml * self.per_sec

                    # while loop for checking evry 0.3 seconds
                    while True:
                        # receive water temperature from arduino
                        sleep(0.3)

                        T_body, T_water = self.get_temp()
                        if T_body or T_water is None :
                            raise Exception

                        current_temp = T_body

                        # put the body temperature in Temp class to compute PID control result
                        control_output = T.compute(current_temp)
                        # check to turn on  or turned off heater

                        if control_output >= 0:
                            GPIO.output(self.pin_heater, False)
                        else:
                            GPIO.output(self.pin_heater, True)

                        GPIO.output(self.pin_pump, True)
                        if T_body >= temp-30 and per :
                            GPIO.output(self.pin_pump, False)
                            sleep(1)
                            GPIO.output(self.pin_pump, True)
                            per = False
                        if T_body >= temp :
                            if 0 < t < 1:
                                GPIO.output(self.pin_pump, False)
                            elif 2 < t < 3:
                                GPIO.output(self.pin_pump, False)
                            elif 4 < t < 5:
                                GPIO.output(self.pin_pump, False)
                            elif 6 < t < 7:
                                GPIO.output(self.pin_pump, False)
                            elif 8 < t < 9:
                                GPIO.output(self.pin_pump, False)
                            elif 10 < t < 11:
                                GPIO.output(self.pin_pump, False)
                            elif 12 < t < 13:
                                GPIO.output(self.pin_pump, False)
                            elif 14 < t < 15:
                                GPIO.output(self.pin_pump, False)
                            elif 16 < t < 17:
                                GPIO.output(self.pin_pump, False)
                            elif 18 < t < 19:
                                GPIO.output(self.pin_pump, False)
                            elif 20 < t < 21:
                                GPIO.output(self.pin_pump, False)
                            elif 22 < t < 23:
                                GPIO.output(self.pin_pump, False)
                            elif 24 < t < 25:
                                GPIO.output(self.pin_pump, False)
                            elif 26 < t < 27:
                                GPIO.output(self.pin_pump, False)
                            elif 28 < t < 29:
                                GPIO.output(self.pin_pump, False)
                            elif 30 < t < 31:
                                GPIO.output(self.pin_pump, False)
                            elif 32 < t < 33:
                                GPIO.output(self.pin_pump, False)
                            t += 0.31
                        if t >= time:
                            break

                    # shut down pump and heater
                    GPIO.output(self.pin_pump, True)
                    GPIO.output(self.pin_heater, True)
                self.log_machine.get_logger(self.ID, f"hot milk done!! on temp {T_body}", 1)
            except Exception as e :
                self.off()
                self.error = True
                self.per_machine = False
                self.log_machine.get_logger(self.ID, f"Hot mikk failed : {e}", 3)

    # frothing milk
    def froth_milk(self, temp: int, ml: float):
        global T_water
        global T_body
        global per
        per = True
        if self.per_machine:
            try:
                if ml > 0:
                    self.per_sec = self.pSec['pump']
                    # set temperature given by class Temp
                    T = Temp(kp=1, ki=0.1, kd=0.8, set_point=temp + 20, simple_time=0.31)

                    # set some statics variables
                    t = 0
                    sleep(2)
                    time = ml * self.per_sec

                    # while loop for checking evry 0.3 seconds
                    while True:
                        # receive water temperature from arduino
                        sleep(0.3)

                        T_body, T_water = self.get_temp()
                        if T_body or T_water is None:
                            raise Exception

                        current_temp = T_body

                        # put the body temperature in Temp class to compute PID control result
                        control_output = T.compute(current_temp)
                        # check to turn on  or turned off heater

                        if control_output >= 0:
                            GPIO.output(self.pin_heater, False)
                        else:
                            GPIO.output(self.pin_heater, True)

                        GPIO.output(self.pin_pump, True)
                        if T_body >= temp - 30 and per:
                            GPIO.output(self.pin_pump, False)
                            sleep(1)
                            GPIO.output(self.pin_pump, True)
                            per = False
                        if T_body >= temp:
                            if 0 < t < 1:
                                GPIO.output(self.pin_pump, False)
                            elif 2 < t < 3:
                                GPIO.output(self.pin_pump, False)
                            elif 4 < t < 5:
                                GPIO.output(self.pin_pump, False)
                            elif 6 < t < 7:
                                GPIO.output(self.pin_pump, False)
                            elif 8 < t < 9:
                                GPIO.output(self.pin_pump, False)
                            elif 10 < t < 11:
                                GPIO.output(self.pin_pump, False)
                            elif 12 < t < 13:
                                GPIO.output(self.pin_pump, False)
                            elif 14 < t < 15:
                                GPIO.output(self.pin_pump, False)
                            elif 16 < t < 17:
                                GPIO.output(self.pin_pump, False)
                            elif 18 < t < 19:
                                GPIO.output(self.pin_pump, False)
                            elif 20 < t < 21:
                                GPIO.output(self.pin_pump, False)
                            elif 22 < t < 23:
                                GPIO.output(self.pin_pump, False)
                            elif 24 < t < 25:
                                GPIO.output(self.pin_pump, False)
                            elif 26 < t < 27:
                                GPIO.output(self.pin_pump, False)
                            elif 28 < t < 29:
                                GPIO.output(self.pin_pump, False)
                            elif 30 < t < 31:
                                GPIO.output(self.pin_pump, False)
                            elif 32 < t < 33:
                                GPIO.output(self.pin_pump, False)
                            t += 0.31
                        if t >= time:
                            break

                    # shut down pump and heater
                    GPIO.output(self.pin_pump, True)
                    GPIO.output(self.pin_heater, True)
                self.log_machine.get_logger(self.ID, f"froth milk done!! on temp {T_body}", 1)
            except Exception as e:
                self.off()
                self.error = True
                self.per_machine = False
                self.log_machine.get_logger(self.ID, f"froth mikk failed : {e}", 3)

    # heat air
    def heater(self, temp: int, time: float ):
        global T_water
        global T_body
        if self.per_machine:
            try :
                if time > 0 :
                    # set temperature given by class Temp
                    T = Temp(kp=1, ki=0.1, kd=0.8, set_point=temp+10, simple_time=0.31)

                    # set some statics variables
                    t = 0
                    sleep(2)
                    # while loop for checking evry 0.3 seconds
                    while True:
                        # receive water temperature from arduino
                        sleep(0.3)

                        T_body, T_water = self.get_temp()
                        if T_body or T_water is None:
                            raise Exception

                        current_temp = T_body

                        # put the body temperature in Temp class to compute PID control result
                        control_output = T.compute(current_temp)
                        # check to turn on  or turned off heater

                        if control_output >= 0:
                            GPIO.output(self.pin_heater, False)
                        else:
                            GPIO.output(self.pin_heater, True)

                        if T_body >= temp :
                            t += 0.31

                        if t > time:
                            break
                    # shut down pump and heater
                    GPIO.output(self.pin_pump, True)
                    GPIO.output(self.pin_heater, True)
                self.log_machine.get_logger(self.ID, f"heater done!! on {T_body}",1)
            except Exception as e :
                self.off()
                self.error = True
                self.per_machine = False
                self.log_machine.get_logger(self.ID, f"heater failed : {e}", 3)

    # heating water
    def water(self, sec: float):
        if self.per_machine:
            try :
                if sec > 0 :
                    # set some statics variables
                    GPIO.output(self.pin_pump, False)
                    sleep(sec)
                    GPIO.output(self.pin_pump, True)

                self.log_machine.get_logger(self.ID, f"water Done!!", 1)
            except Exception as e :
                self.off()
                self.error = True
                self.per_machine = False
                self.log_machine.get_logger(self.ID, f"water failed : {e}", 3)




    # for grinder
    def grinder(self, gram: float):
        if self.per_machine:
            try :
                if gram > 0 :
                    pin = self.oPin['pinGrinder']
                    per_sec = self.pSec['grinder']
                    t = per_sec * gram
                    sleep(1)
                    GPIO.output(pin, False)
                    sleep(t)
                    GPIO.output(pin, True)
                sleep(0.5)
                self.log_machine.get_logger(self.ID, f"grinder done!! //gram : {gram}", 1)
            except Exception as e:
                self.off()
                self.error = True
                self.per_machine = False
                self.log_machine.get_logger(self.ID, f"grinder failed : {e}", 3)

    # for silos
    # number can be 1 or 2 or 3 (Right to left)
    def silo(self, gram: float, number: int):
        if self.per_machine:
            try :
                if gram > 0 :
                    number = str(number)
                    pin = self.oPin['pinSilo'+number]
                    per_sec = self.pSec['Silo'+number]
                    t = per_sec * gram
                    sleep(1)
                    GPIO.output(pin, False)
                    sleep(t)
                    GPIO.output(pin, True)

                self.log_machine.get_logger(self.ID, f"silo {number} done!! // gram : {gram}", 1)
            except Exception as e:
                self.off()
                self.error = True
                self.per_machine = False
                self.log_machine.get_logger(self.ID, f"silo {number} failed : {e}", 3)

    # for mixer
    def mixer(self, state: int):    # 1 for on and 0 for off
        if self.per_machine:
            try :
                pin = self.oPin['pinMixer']
                sleep(0.5)
                if state == 1 :
                    GPIO.output(pin, False)
                elif state == 0 :
                    GPIO.output(pin, True)
                self.log_machine.get_logger(self.ID, f"mixer done!!//", 1)
            except Exception as e:
                self.off()
                self.error = True
                self.per_machine = False
                self.log_machine.get_logger(self.ID, f"Mixer failed : {e}", 3)

    # for mixer
    def mix_valve(self, state: int):  # 1 for on and 0 for off
        if self.per_machine:
            try:
                pin = self.oPin['pinMixValve']
                sleep(0.5)
                if state == 1:
                    GPIO.output(pin, False)
                elif state == 0:
                    GPIO.output(pin, True)
                self.log_machine.get_logger(self.ID, f"Valve opened or closed!!", 1)
            except Exception as e:
                self.off()
                self.error = True
                self.per_machine = False
                self.log_machine.get_logger(self.ID, f"Mixer Valve failed : {e}", 3)

    # for cup
    def dispenser(self):
        pin = self.oPin['pinDispenser']
        sen = self.iPin['senDispenser']
        r_sec = self.pSec['dispenser']
        sleep(1)
        GPIO.output(pin, False)
        sleep(r_sec)
        GPIO.output(pin, True)
        # TODO: add sensor codes

    # for move nozzle
    def nozzle(self, state):                                                 # state can be 0 or 1 (0 means off amd 1 means on)
        pinOn = self.oPin['pinNozzleOn']
        pinOff = self.oPin['pinNozzleOff']
        r_sec = self.pSec['nozzle']
        len = self.pSec['lenNozzle']
        t = r_sec * len
        if state == 0 :
            sleep(1)
            GPIO.output(pinOff, False)
            sleep(t)
            GPIO.output(pinOff, True)
        if state == 1:
            sleep(1)
            GPIO.output(pinOn, False)
            sleep(t)
            GPIO.output(pinOn, True)
        GPIO.output(pinOn, True)
        GPIO.output(pinOff, True)

    # for open and close exit gate
    def gate(self, state): # state cna be 0 or 1 (0 means close amd 1 means open)
        pinOpen = self.oPin['pinOpen']
        pinClose = self.oPin['pinClose']
        # TODO: copy codes from RasPi

    # verify for open the door
    def verify(self):
        # TODO: connect RFID-RC522 to raspberry PI or arduino
        # TODO: To identify and open door
        pass

    # check cards for validity discount
    def free_coffee(self):
        # TODO: connect RFID-RC522 to raspberry PI or arduino
        # TODO: To give discount price
        pass

    # function as Run elements of machine by sorting
    def menu(self):

        if self.coffee_powder > 12 :
            times = 2
            self.modCoffeePowder = self.coffee_powder/2
            self.modCoffee = self.coffee / 2
        elif self.coffee_powder == 0 :
            times = 0
        else :
            times = 1
            self.modCoffeePowder = self.coffee_powder
            self.modCoffee = self.coffee


        # for loop for run coffee cycle
        for i in range(times) :
            # set valve position
            self.valve(0)
            self.water(5)
            self.valve(10)
            # check deriver is home
            self.derive(0)
            # drop cup
            # self.dispenser()
            # coffee powder
            self.grinder(self.modCoffeePowder)
            # set drive position to work
            self.derive(1)
            self.water(1.5)
            # set nozzle position
            #self.nozzle(1)
            # run hot water
            self.hot_water(120, 130, self.modCoffee)
            sleep(10)
            # move one step of valve
            self.valve(10)
            sleep(10)
            # drive back to home
            self.derive(0)


        # sugar powder
        # self.silo(self.sugar, 1)
        # chocolate powder
        # self.silo(self.chocolate_powder, 2)
        # cocoa powder
        # self.silo(self.cocoa_powder, 3)


        # setup for hot milk cycle
        #if self.hotMilk > 0 :
            # set valve position before hot state 3
            #self.valve(2)
            # warm water in heater
            #self.heater(130, 0)
            # set valve position to froth state 5
            #self.valve(10)
            # mix on
            # self.mixer(1)
            # hot milk setup
            #self.hot_milk(135, self.hotMilk)
            # mix off
            # self.mixer(0)
            # move one step of valve
            #self.valve(40)
            # heater on
            #self.heater(120, 0)
            # open and close mixer valve
            # self.mix_valve(1)
            # sleep(5)
            # self.mix_valve(0)


        # setup for froth milk cycle
        if self.frothMilk > 0 :
            # set valve position before froth state 5
            self.valve(5)
            # warm water in heater
            #self.heater(130, 0)
            # set valve position to froth state 5
            #self.valve(10)
            # mix on
            # self.mixer(1)
            # froth milk setup
            self.froth_milk(130, self.frothMilk)
            sleep(0)
            # mix off
            # self.mixer(0)
            # move one step of valve
            self.valve(20)
            # heater on
            self.froth_milk(130, 10)
            # open and close mixer valve
            self.mix_valve(1)
            sleep(5)
            self.mix_valve(0)


        self.valve(5)
        # mix on
        # self.mixer(1)
        self.water(10)
        # mix off
        # self.mixer(0)
        self.valve(10)
        # nozzle back
        #self.nozzle(0)
        # open the gate
        #self.gate(1)
        #sleep(30)
        # close the gate
        #self.gate(0)
        # clean mode
        # setup valve to clean mod
        #self.valve(3)
        # milk back to tank
        #self.heater(100, 10)
        # move one step to evacuation valve
        #self.valve(10)
